searchData={"items":[{"type":"task","title":"mix sorcery.init","doc":"","ref":"Mix.Tasks.Sorcery.Init.html"},{"type":"function","title":"Mix.Tasks.Sorcery.Init.run/1","doc":"","ref":"Mix.Tasks.Sorcery.Init.html#run/1"},{"type":"module","title":"Sorcery","doc":"To get started with Sorcery, let's use the generator\n\n`mix sorcery.init`\n\nBy default it creates a new namespace 'Src' in your app.\nThe src.ex module is very special, it starts up some :ets/:mnesia tables, as well as holds together some config for making everything else work.\n\nYou can manually create your own with \n```elixir\ndefmodule MyApp.Src do\n  use Sorcery\nend\n```\nAnd then make sure your queries and schemas, etc. are all namespaced below that like `MyApp.Src.Queries.GetStuff` and 'MyApp.Src.Schemas.Thing'","ref":"Sorcery.html"},{"type":"behaviour","title":"Sorcery.GenServerHelpers","doc":"Functions for combining Sorcery and GenServer clients.\nPut these in your init function to create portals and receive updates.","ref":"Sorcery.GenServerHelpers.html"},{"type":"callback","title":"Sorcery.GenServerHelpers.initialize_sorcery/2","doc":"Puts a :sorcery key in state, with a bunch of stuff used behind the scenes by Sorcery.\n\nThis is mandatory, before you can spawn any portals\n\nYou must pass in the sorcery_module. If you used the generator, it'll just be Src.","ref":"Sorcery.GenServerHelpers.html#c:initialize_sorcery/2"},{"type":"callback","title":"Examples - Sorcery.GenServerHelpers.initialize_sorcery/2","doc":"iex> body = %{sorcery_module: Src}\n    iex> state = initialize_sorcery(state, body)","ref":"Sorcery.GenServerHelpers.html#c:initialize_sorcery/2-examples"},{"type":"callback","title":"Sorcery.GenServerHelpers.portal_view/2","doc":"Get a list of entities at the given lvar. \n\nIf this lvar only exists in ONE portal, you can leave out the portal_name, just know that it might be imperceptibly slower since the function needs to iterate over all portals until it finds one with that lvar..\n\n```elixir\n<% players = portal_view(@sorcery, \"?all_players\") %>\n```","ref":"Sorcery.GenServerHelpers.html#c:portal_view/2"},{"type":"callback","title":"Sorcery.GenServerHelpers.portal_view/3","doc":"Get a list of entities at the given portal/lvar\nYou probably want to use this inside heex\n```elixir\n<% players = portal_view(@sorcery, :my_portal, \"?all_players\") %>\n```","ref":"Sorcery.GenServerHelpers.html#c:portal_view/3"},{"type":"callback","title":"Sorcery.GenServerHelpers.spawn_portal/2","doc":"Creates a portal between the GenServer and the PortalServer, using the given arguments.\n\nTakes the state, returns it unchanged. An update should happen in a few milliseconds to add more data to the state..","ref":"Sorcery.GenServerHelpers.html#c:spawn_portal/2"},{"type":"callback","title":"Examples - Sorcery.GenServerHelpers.spawn_portal/2","doc":"iex> body = %{portal_server: Postgres, portal_name: :my_portal, query_module: MyQuery, query_args: %{player_id: 1}}\n    iex> state = spawn_portal(state, body)","ref":"Sorcery.GenServerHelpers.html#c:spawn_portal/2-examples"},{"type":"module","title":"Sorcery.Killbot","doc":"This is a garbage collector, of sorts.\n\nEvery time it runs, it will scan SorceryDb for all dead pids and hang onto those until the postmortem_delay is over\nFor every pid in which the delay IS over, it will find the entities in their portals, filter out any entities that are being watched in other portals, and then remove them from :mnesia.\n```elixir\ndefmodule MyApp.Sorcery do\n  use Sorcery,\n    killbot: %{\n      # Defaults to 10 minutes\n      interval: 10 * 60 * 1_000, \n\n      # Wait 2 intervals AFTER the pid dies, before wiping the entities       \n      # It is likely a user will close a page and come back a few minutes later...\n      # So we don't want to uncache things too quickly.\n      #\n      # Again, I stress that this is NOT milliseconds, but the number of intervals.\n      postmortem_delay: 2        \n    }\nend\n```","ref":"Sorcery.Killbot.html"},{"type":"function","title":"Sorcery.Killbot.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Sorcery.Killbot.html#child_spec/1"},{"type":"function","title":"Sorcery.Killbot.get_watched/0","doc":"Return all the dead portals being watched, which have not yet been removed.","ref":"Sorcery.Killbot.html#get_watched/0"},{"type":"function","title":"Sorcery.Killbot.run_now/0","doc":"Start a run immediately. This does speed up the postmortem_delay","ref":"Sorcery.Killbot.html#run_now/0"},{"type":"behaviour","title":"Sorcery.LiveHelpers","doc":"Functions for combining Sorcery and LiveViews","ref":"Sorcery.LiveHelpers.html"},{"type":"callback","title":"Sorcery.LiveHelpers.initialize_sorcery/2","doc":"Puts a :sorcery key in socket.assigns, with a bunch of stuff used behind the scenes by Sorcery.\n\nThis is mandatory, before you can spawn any portals\n\nYou must pass in the sorcery_module. If you used the generator, it'll just be Src.","ref":"Sorcery.LiveHelpers.html#c:initialize_sorcery/2"},{"type":"callback","title":"Examples - Sorcery.LiveHelpers.initialize_sorcery/2","doc":"iex> body = %{sorcery_module: Src}\n    iex> socket = initialize_sorcery(socket, body)","ref":"Sorcery.LiveHelpers.html#c:initialize_sorcery/2-examples"},{"type":"callback","title":"Sorcery.LiveHelpers.portal_view/2","doc":"Get a list of entities at the given lvar. \n\nIf this lvar only exists in ONE portal, you can leave out the portal_name, just know that it might be imperceptibly slower since the function needs to iterate over all portals until it finds one with that lvar..\n\n```elixir\n<% players = portal_view(@sorcery, \"?all_players\") %>\n```","ref":"Sorcery.LiveHelpers.html#c:portal_view/2"},{"type":"callback","title":"Sorcery.LiveHelpers.portal_view/3","doc":"Get a list of entities at the given portal/lvar\nYou probably want to use this inside heex\n```elixir\n<% players = portal_view(@sorcery, :my_portal, \"?all_players\") %>\n```","ref":"Sorcery.LiveHelpers.html#c:portal_view/3"},{"type":"callback","title":"Sorcery.LiveHelpers.spawn_portal/3","doc":"Creates a portal between the LiveView and the PortalServer, using the given arguments.\n\nTakes the socket, returns it unchanged.","ref":"Sorcery.LiveHelpers.html#c:spawn_portal/3"},{"type":"callback","title":"Examples - Sorcery.LiveHelpers.spawn_portal/3","doc":"iex> body = %{portal_server: Postgres, portal_name: :my_portal, query_module: MyQuery, query_args: %{player_id: 1}}\n    iex> socket = spawn_portal(socket, body)","ref":"Sorcery.LiveHelpers.html#c:spawn_portal/3-examples"},{"type":"type","title":"Sorcery.LiveHelpers.socket_type/0","doc":"","ref":"Sorcery.LiveHelpers.html#t:socket_type/0"},{"type":"module","title":"Sorcery.Mutation","doc":"This is the interface for creating changes in Sorcery.\nMutations have some similarities to Ecto Changesets in the sense that we are building a struct that defines all the changes without actually applying them.\nBut there are many differences, as you will soon see.\n\nYou start by creating the initial mutation with a portal.\n\nThen you apply all the changes you want to the data in that portal.\n\nAt first nothing happens! This is because you must send the mutation up to the parent PortalServer.\n\nAfter it applies the changes, it will dispatch them back down to any child PortalServers that care about any of the entities involved.\nThis is where the magic of Sorcery really comes in. Under the hood, we are doing some pretty efficient reverse queries, in parallel, to figure out where to send the data.\n\nAll the PortalServers will update their Portals accordingly.\n\n```elixir\niex> alias Sorcery.Mutation, as: M\niex> m = M.init(socket.assigns.sorcery, :my_portal)\niex> m = M.update(m, [:player, 1, :age], fn _original_age, latest_age -> latest_age + 1 end)\niex> m = M.put(m, [:player, 1, :health], 100)\niex>","ref":"Sorcery.Mutation.html"},{"type":"module","title":"We can also use placeholder ids so that new entities (which haven't even been created yet) can be referenced by other entities. - Sorcery.Mutation","doc":"iex> m = M.create_entity(m, :team, \"?my_new_team\", %{name: \"My New Team\"})\niex> m = M.put(m, [:player, 1, :team_id], \"?my_new_team.id\")\niex> # This funny string is of course, the placeholder. But after we actually run the mutation, the team is created with a normal integer id.\niex> # The parent PortalServer creates the new team entity, it will automatically replace all calls to \"?my_new_team\" with that entity.\niex> Sorcery.Mutation.send_mutation(m)\niex> # ... a few milliseconds later, after receiving the update\niex> player = portal_view(@sorcery, :my_portal, \"?all_players\")[1]\niex> new_team_id = player.team_id\niex> is_integer(new_team_id)\ntrue\niex> team = portal_view(@sorcery, :my_portal, \"?all_teams\")[new_team_id]\niex> team.name\n\"My New Team\"\n```","ref":"Sorcery.Mutation.html#module-we-can-also-use-placeholder-ids-so-that-new-entities-which-haven-t-even-been-created-yet-can-be-referenced-by-other-entities"},{"type":"function","title":"Sorcery.Mutation.create_entity/4","doc":"One of the benefits of a Mutation is that you can create entities using placeholder ids.","ref":"Sorcery.Mutation.html#create_entity/4"},{"type":"function","title":"Examples - Sorcery.Mutation.create_entity/4","doc":"iex> m = M.create_entity(m, :team, \"?my_new_team\", %{name: \"My New Team\"})\n    iex> m = M.put(m, [:player, 1, :team_id], \"?my_new_team\")\n    iex> # It always defaults to using the :id, but you can specify another field\n    iex> m = M.put(m, [:player, 1, :team_id], \"?my_new_team.my_field\")","ref":"Sorcery.Mutation.html#create_entity/4-examples"},{"type":"function","title":"Sorcery.Mutation.delete_entity/3","doc":"Sometimes we want to delete an entity entirely. Be careful, after the mutation is run, this cannot be undone","ref":"Sorcery.Mutation.html#delete_entity/3"},{"type":"function","title":"Sorcery.Mutation.get_in/2","doc":"","ref":"Sorcery.Mutation.html#get_in/2"},{"type":"function","title":"Sorcery.Mutation.init/2","doc":"The first step of every mutation. You must have an existing portal in order to do any of this.","ref":"Sorcery.Mutation.html#init/2"},{"type":"function","title":"Examples - Sorcery.Mutation.init/2","doc":"iex> m = Sorcery.Mutation.init(socket.assigns.sorcery, :my_portal)\n    iex> is_struct(m)\n    true","ref":"Sorcery.Mutation.html#init/2-examples"},{"type":"function","title":"Sorcery.Mutation.put/3","doc":"Just like put_in, but for mutations.","ref":"Sorcery.Mutation.html#put/3"},{"type":"function","title":"Examples - Sorcery.Mutation.put/3","doc":"iex> m = Sorcery.Mutation.put(m, [:player, 1, :health], 100)","ref":"Sorcery.Mutation.html#put/3-examples"},{"type":"function","title":"Sorcery.Mutation.send_mutation/2","doc":"Sends the mutation to the corresponding Portal Server\nThe Portal Server will then update its own data store.\nAll portals that care about the changes will automatically update accordingly (including the one calling this function) \n\nReturns the mutation passed in.\n\nThe return state will include a temp_portal, which is handy for both testing, and for optimistic updates. \nWhen the PortalServer sends the new, fully updated data, then it will overwrite the portal and remove the temp_portal data.\nThere are limitations to the temp_portal, and it should not be trusted too much.\n\nTakes an optional keyword list of options. Available options: :optimistic, :handle_fail, :handle_success\n\n### :optimistic (true)\nWhen true, this will attempt an optimistic update. Does not work great when creating a new entity because we do not yet have the id.\n\n### :handle_fail (nil)\nExpects nil, or a callback function of the shape `fn error, state -> state end`\nFor example if you use Mutation.validate, which fails, then the error argument will be a map including keys :kind, :reason\n\n### :handle_success (nil)\nExpects nil, or a callback function of the shape `fn data, state -> state end`\nDO NOT try to manually persist the data into a portal. That happens automatically after SorceryDb does some work.\nInstead this can be useful as a sanity test, or to verify that the transaction has, indeed, completed successfully.\n\nIf you are using Sorcery.LiveHelpers, this takes a socket instead of state, allowing you to redirect, put_flash, etc.","ref":"Sorcery.Mutation.html#send_mutation/2"},{"type":"function","title":"Sorcery.Mutation.send_mutation/3","doc":"","ref":"Sorcery.Mutation.html#send_mutation/3"},{"type":"function","title":"Sorcery.Mutation.skip/2","doc":"Mark a mutation to NOT be sent to the PortalServer.\nInstead such a mutation will return an error tuple.\nWhen using the LiveView helper with optimistic_mutation/2, skipped mutations will trigger put_flash/3","ref":"Sorcery.Mutation.html#skip/2"},{"type":"function","title":"Examples - Sorcery.Mutation.skip/2","doc":"iex> mutation = skip(mutation, :info, \"Testing\")\n    iex> send_mutation(mutation, sorcery)\n    {:error, {:skip, :info, \"Testing\"}}","ref":"Sorcery.Mutation.html#skip/2-examples"},{"type":"function","title":"Sorcery.Mutation.skip/3","doc":"","ref":"Sorcery.Mutation.html#skip/3"},{"type":"function","title":"Sorcery.Mutation.update/3","doc":"Kind of like a fusion between Kernel.update_in/3 and Map.update/4\n\nYou need to pass in both a path list, and a default value in case the original value is nil.","ref":"Sorcery.Mutation.html#update/3"},{"type":"function","title":"Examples - Sorcery.Mutation.update/3","doc":"iex> Sorcery.Mutation.update(m, [:player, 1, :age], 100, fn _original_age, latest_age -> latest_age + 1 end)","ref":"Sorcery.Mutation.html#update/3-examples"},{"type":"function","title":"Sorcery.Mutation.validate/3","doc":"The function must return either: \n:ok\nor\n{kind, reason}\n\nfor example\n```elixir\n    {:error, \"You can't do that.\"}\n```\n\nThis is used by LiveHelpers.optimistic_mutation to skip mutations and show flash messages instead.\n\nThe function passed in takes two arguments, old_data and new_data\nBe careful, the new_data is not real. It is simply the best guess based on the current diff. By time the mutation reaches the portal server, it could be different!","ref":"Sorcery.Mutation.html#validate/3"},{"type":"function","title":"Examples - Sorcery.Mutation.validate/3","doc":"iex> Mutation.validate(mutation, [:tk, :id, :some_attr], fn v -> \n      if v > 0, do: :ok, else: {:error, \"Something went wrong\"}\n    end)","ref":"Sorcery.Mutation.html#validate/3-examples"},{"type":"module","title":"Sorcery.Mutation.Temp","doc":"","ref":"Sorcery.Mutation.Temp.html"},{"type":"function","title":"Sorcery.Mutation.Temp.add_temp_portal/2","doc":"","ref":"Sorcery.Mutation.Temp.html#add_temp_portal/2"},{"type":"function","title":"Sorcery.Mutation.Temp.get_split_data/1","doc":"","ref":"Sorcery.Mutation.Temp.html#get_split_data/1"},{"type":"module","title":"Sorcery.PortalServer","doc":"A PortalServer is a special type of GenServer. Portals are created between different PortalServers, always in a hierarchical parent <--> child relationship.\n\nEvery PortalServer must have access to some kind of data store. That might be Postgres via Ecto.Repo... or it could be a LiveView storing data in socket.assigns. (Remember, LiveViews *are* GenServers!)\nYou can even create your own adapter and use ANY backend as the data store, as long as you can query and mutate the data.","ref":"Sorcery.PortalServer.html"},{"type":"module","title":"Setup - Sorcery.PortalServer","doc":"A PortalServer is any GenServer that holds a special :sorcery key somewhere in its state, and also handles messages sent to {:sorcery, msg}","ref":"Sorcery.PortalServer.html#module-setup"},{"type":"function","title":"Sorcery.PortalServer.add_portal_server_state/2","doc":"To convert any process into a PortalServer, you must follow 2 steps:\n1) Add some PortalServer configuration to the state\n2) Implement the PortalServer handlers for a handle_info function\n\n\n```elixir\nuse GenServer\nuse Sorcery.PortalServer\n\ndef init(_) do\n  state = %{} # You can still add whatever you want here\n\n  state = Sorcery.PortalServer.add_portal_server_state(state, %{\n    config_module: Src,\n    store_adapter: Sorcery.StoreAdapters.Ecto,\n\n    # This depends on the adapters you use\n    args: %{\n      repo_module: MyApp.Repo\n    }\n  })\n\n  {:ok, state}\nend\n```","ref":"Sorcery.PortalServer.html#add_portal_server_state/2"},{"type":"module","title":"Sorcery.PortalServer.InnerState","doc":"This exists to resolve a lot of confusion caused by generically using the word 'state' to refer to either the data inside the :sorcery key, or the data *including* :sorcery.\nFor example a LiveView would have socket.assigns.sorcery\nSo does state refer to socket.assigns? Or socket.assigns.sorcery?\nIn theory, Sorcery should never need to access the outer map, aside from maybe the helpers at most.\n\nSo to save confusion, InnerState is ONLY ever the value of the :sorcery key.\nThere is no OuterState struct, as it depends on how the user configures the PortalServer.","ref":"Sorcery.PortalServer.InnerState.html"},{"type":"function","title":"Sorcery.PortalServer.InnerState.new/1","doc":"","ref":"Sorcery.PortalServer.InnerState.html#new/1"},{"type":"module","title":"Sorcery.PortalServer.Portal","doc":"A Portal represents an ongoing SrcQL Query.\nRather than grabbing data and ending, it continues watching for changes to the results set.\n\nA copy of any given Portal is found in the state of two different PortalServers; the parent, and the child PortalServer.","ref":"Sorcery.PortalServer.Portal.html"},{"type":"function","title":"Sorcery.PortalServer.Portal.all_lvars_by_tk/2","doc":"Get the list of all lvars that match a given tk","ref":"Sorcery.PortalServer.Portal.html#all_lvars_by_tk/2"},{"type":"function","title":"Sorcery.PortalServer.Portal.freeze/1","doc":"When we freeze a portal, all the lvars are replaced with tablekeys (tk).\nIf multiple lvars reference the same tablekey, they are merged.\n\nA frozen portal is no longer kept up to date. Instead, it is immutable.","ref":"Sorcery.PortalServer.Portal.html#freeze/1"},{"type":"function","title":"Sorcery.PortalServer.Portal.get_in/3","doc":"","ref":"Sorcery.PortalServer.Portal.html#get_in/3"},{"type":"function","title":"Sorcery.PortalServer.Portal.new/1","doc":"","ref":"Sorcery.PortalServer.Portal.html#new/1"},{"type":"behaviour","title":"Sorcery.Query","doc":"A query module defines, in plain elixir data structures, the kind of data we want to watch for.\n\nThe syntax takes some inspiration from Datalog, but with many differences as well.\n\n```elixir\ndefmodule Src.Queries.GetBattle do\n\n  use Sorcery.Query, %{\n\n    # Args will be passed in later when the query is called.\n    args: %{\n      player_id: :integer\n    },\n\n    # This is the meat of any query. Read it one row at a time.\n    # Every row has either 3, 4, columns\n    where: [\n      # 4 column syntax:\n      # [lvar,        tk,            attr,         value]\n      #\n      #\n      # So we start with an lvar (or 'Logic Variable') called \"?player\"\n      # It represents a set of entities with the Schema of :player\n      # And we are filtering them such that ?player.id == args.player_id\n      # The arg MUST be declared in the args map.\n      [ \"?player\",    :player,       :id,          :args_player_id],\n\n      # Now we make a new lvar called \"?team\"\n      # See how it now references the previous lvar?\n      # So we are filtering all teams such that team.id matches ANY ?player.team_id\n      [ \"?team\",      :team,         :id,          \"?player.team_id\"],\n      [ \"?arena\",     :battle_arena, :id,          \"?team.location_id\"],\n\n      # This is not the same as ?team.\n      # Its a new lvar using the same schema, but with a different set of filters\n      # So we're getting all teams such that team.location_id == ?arena.id\n      [ \"?all_teams\", :team,         :location_id, \"?arena.id\"],\n\n      # Now we use the 3 column syntax, just to avoid repetition.\n      # This could also be rewritten as two rows:\n      # [\"?all_players\", :player, :team_id, \"?all_teams.id\"],\n      # [\"?all_players\", :player, :health, {:>, 0}],\n      [ \"?all_players\", :player, [\n        {:team_id, \"?all_teams.id\"},\n        {:health, {:>, 0}},\n      ]],\n      # Notice the value above {:>, 0}\n      # By default, every value automatically expands under the hood to {:==, value}\n      # But if you want to manually use an operator, you can.\n      # Possible operators: :==, :!=, :>, :>=, :<, :<=, :in\n\n\n      [ \"?spells\", :spell_instance, :player_id, \"?all_players.id\"],\n      [ \"?spell_types\", :spell_type, :id, \"?spells.type_id\"],\n    ],\n\n    # Without a find map, the query returns no results.\n    # We do not necessarily need all the lvars, nor all the fields\n    # If we want all available fields, use :*\n    # Otherwise pass in a list of specific ones. The :id attr is automatically added.\n    find: %{\n      \"?arena\" => :*,\n      \"?all_teams\" => [:name, :location_id],\n      \"?all_players\" => :*,\n      \"?spells\" => :*,\n      \"?spell_types\" => :*,\n    }\n  }\n\nend\n\n```","ref":"Sorcery.Query.html"},{"type":"callback","title":"Sorcery.Query.clauses/1","doc":"Return a list of WhereClause structs for the Query module","ref":"Sorcery.Query.html#c:clauses/1"},{"type":"callback","title":"Examples - Sorcery.Query.clauses/1","doc":"iex> [clause1 | _] = Src.Queries.GetBattle.clauses(%{player_id: 1})\n    iex> clause1\n    %Sorcery.Query.WhereClause{lvar: :\"?player\", tk: :player, attr: :id, left: nil, right: 1, op: :==, other_lvar: nil, other_lvar_attr: nil, arg_name: nil, right_type: :literal}","ref":"Sorcery.Query.html#c:clauses/1-examples"},{"type":"callback","title":"Sorcery.Query.finds/0","doc":"Returns the finds. This differs from raw_struct().finds because :id fields have been added.","ref":"Sorcery.Query.html#c:finds/0"},{"type":"callback","title":"Examples - Sorcery.Query.finds/0","doc":"iex> Src.Queries.GetBattle.finds()\n    %{\"?arena\": :*, \"?all_teams\": [:location_id, :id, :name], \"?all_players\": :*, \"?spells\": :*, \"?spell_types\": :*}","ref":"Sorcery.Query.html#c:finds/0-examples"},{"type":"function","title":"Sorcery.Query.from_tk_map/3","doc":"If you have a map in the format of `%{tk => %{id => %{...entity...}}}`\nThen you can use it like a database and run the query against it.","ref":"Sorcery.Query.html#from_tk_map/3"},{"type":"callback","title":"Sorcery.Query.raw_struct/0","doc":"Returns a Sorcery.Query struct.","ref":"Sorcery.Query.html#c:raw_struct/0"},{"type":"callback","title":"Examples - Sorcery.Query.raw_struct/0","doc":"iex> q = Src.Queries.GetBattle.raw_struct()\n    iex> is_struct(q, Sorcery.Query)\n    iex> q.find[\"?all_players\"]\n    :*\n    iex> [clause1 | _] = q.where\n    iex> clause1\n    [\"?player\", :player, :id, :args_player_id]\n    iex> Enum.at(q.lvar_tks, 3)\n    {\"?all_teams\", :team}","ref":"Sorcery.Query.html#c:raw_struct/0-examples"},{"type":"callback","title":"Sorcery.Query.tks_affected/0","doc":"A unique list of all tks mentioned by this query.","ref":"Sorcery.Query.html#c:tks_affected/0"},{"type":"callback","title":"Examples - Sorcery.Query.tks_affected/0","doc":"iex> Src.Queries.GetBattle.tks_affected()\n    [:player, :team, :battle_arena, :spell_instance, :spell_type]","ref":"Sorcery.Query.html#c:tks_affected/0-examples"},{"type":"type","title":"Sorcery.Query.t/0","doc":"","ref":"Sorcery.Query.html#t:t/0"},{"type":"module","title":"Sorcery.ReturnedEntities","doc":"This is the return type of any forward query, and any generated data.\n\nTry not to access the data directly, but rather use the functions in this module.\n\nThe reason is that Sorcery is still a work in progress, the format might change, but the functions should never break their contract's.\n\nThis is also the basic format of portals.\n\nCurrently the :primary_entities field is not being used. The idea was for it to help with ergonomics when a query is only expected to have one result.","ref":"Sorcery.ReturnedEntities.html"},{"type":"function","title":"Sorcery.ReturnedEntities.apply_find_map/2","doc":"Takes the :find key from some SrcQL and uses that to prune the RE.","ref":"Sorcery.ReturnedEntities.html#apply_find_map/2"},{"type":"function","title":"Examples - Sorcery.ReturnedEntities.apply_find_map/2","doc":"iex> re = Sorcery.ReturnedEntities.new()\n    iex> re = put_entities(re, :bird, [ %{id: 1, name: \"A\"}, %{id: 23, name: \"B\"}])\n    iex> re = put_entities(re, :spell, [ %{id: 1, name: \"A\"}, %{id: 23, name: \"B\"}])\n    iex> get_entities(re, :bird)\n    [ %{id: 1, name: \"A\"}, %{id: 23, name: \"B\"} ]\n    iex> re = apply_find_map(re, %{bird: [:id]})\n    iex> get_entities(re, :bird)\n    [ %{id: 1}, %{id: 23} ]","ref":"Sorcery.ReturnedEntities.html#apply_find_map/2-examples"},{"type":"function","title":"Sorcery.ReturnedEntities.assign_lvar_tk/3","doc":"","ref":"Sorcery.ReturnedEntities.html#assign_lvar_tk/3"},{"type":"function","title":"Sorcery.ReturnedEntities.delete_attrs/3","doc":"Deletes attribute keys from every entity of the given tk","ref":"Sorcery.ReturnedEntities.html#delete_attrs/3"},{"type":"function","title":"Examples - Sorcery.ReturnedEntities.delete_attrs/3","doc":"iex> re = Sorcery.ReturnedEntities.new()\n    iex> re = put_entities(re, :player, [ %{id: 1, name: \"A\"}, %{id: 23, name: \"B\"}])\n    iex> re.data.player\n    %{1 => %{id: 1, name: \"A\"}, 23 => %{id: 23, name: \"B\"}}\n    iex> re = delete_attrs(re, :player, [ :name ])\n    iex> re.data.player\n    %{1 => %{id: 1}, 23 => %{id: 23}}","ref":"Sorcery.ReturnedEntities.html#delete_attrs/3-examples"},{"type":"function","title":"Sorcery.ReturnedEntities.delete_entities/3","doc":"Removes all the listed ids of a given tk.","ref":"Sorcery.ReturnedEntities.html#delete_entities/3"},{"type":"function","title":"Examples - Sorcery.ReturnedEntities.delete_entities/3","doc":"iex> re = Sorcery.ReturnedEntities.new()\n    iex> re = put_entities(re, :player, [ %{id: 1, name: \"A\"}, %{id: 23, name: \"B\"}])\n    iex> re.data.player\n    %{1 => %{id: 1, name: \"A\"}, 23 => %{id: 23, name: \"B\"}}\n    iex> re = delete_entities(re, :player, [ 23 ])\n    iex> re.data.player\n    %{1 => %{id: 1, name: \"A\"}}","ref":"Sorcery.ReturnedEntities.html#delete_entities/3-examples"},{"type":"function","title":"Sorcery.ReturnedEntities.get_entities/2","doc":"","ref":"Sorcery.ReturnedEntities.html#get_entities/2"},{"type":"function","title":"Examples - Sorcery.ReturnedEntities.get_entities/2","doc":"iex> re = Sorcery.ReturnedEntities.new()\n    iex> re = put_entities(re, :player, [ %{id: 1, name: \"A\"}, %{id: 23, name: \"B\"}])\n    iex> get_entities(re, :player)\n    [ %{id: 1, name: \"A\"}, %{id: 23, name: \"B\"} ]","ref":"Sorcery.ReturnedEntities.html#get_entities/2-examples"},{"type":"function","title":"Sorcery.ReturnedEntities.merge/1","doc":"","ref":"Sorcery.ReturnedEntities.html#merge/1"},{"type":"function","title":"Sorcery.ReturnedEntities.new/0","doc":"Creates a new, empty RE struct.","ref":"Sorcery.ReturnedEntities.html#new/0"},{"type":"function","title":"Examples - Sorcery.ReturnedEntities.new/0","doc":"iex> new()\n    %Sorcery.ReturnedEntities{ primary_entities: [], data: %{} }","ref":"Sorcery.ReturnedEntities.html#new/0-examples"},{"type":"function","title":"Sorcery.ReturnedEntities.put_entities/3","doc":"","ref":"Sorcery.ReturnedEntities.html#put_entities/3"},{"type":"function","title":"Examples - Sorcery.ReturnedEntities.put_entities/3","doc":"iex> re = Sorcery.ReturnedEntities.new()\n    iex> re = put_entities(re, :player, [ %{id: 1, name: \"A\"}, %{id: 23, name: \"B\"}])\n    iex> re.data.player\n    %{1 => %{id: 1, name: \"A\"}, 23 => %{id: 23, name: \"B\"}}\n    iex> re = put_entities(re, :player, [ %{id: 1, name: \"C\"}, %{id: 42, name: \"B\"}])\n    iex> re.data.player\n    %{1 => %{id: 1, name: \"C\"}, 23 => %{id: 23, name: \"B\"}, 42 => %{id: 42, name: \"B\"}}","ref":"Sorcery.ReturnedEntities.html#put_entities/3-examples"},{"type":"module","title":"Sorcery.Schema","doc":"In sorcery, you must define your entity types as 'Schemas'\n\n\n```elixir\ndefmodule MyApp.Player do\n  use Sorcery.Schema, fields: %{\n    name: %{t: :string, min: 0, max: 45, default: \"Nameless\"},\n    age: %{t: :integer, min: 0, max: 99, optional?: false},\n  }\nend\n```\n\nSee the guide for a comprehensive table of all possible attributes for each :t.\n\nAfter you do the above setup, The MyApp.Player module will have some neat superpowers. Go take a look in iex.","ref":"Sorcery.Schema.html"},{"type":"function","title":"Sorcery.Schema.meta_defaults/1","doc":"","ref":"Sorcery.Schema.html#meta_defaults/1"},{"type":"behaviour","title":"Sorcery.Schema.FieldType","doc":"WIP\ncurrently it is not possible to add your own custom field types. But in the future it should be possible since they all follow this behaviour","ref":"Sorcery.Schema.FieldType.html"},{"type":"function","title":"Sorcery.Schema.FieldType.add_optional/2","doc":"","ref":"Sorcery.Schema.FieldType.html#add_optional/2"},{"type":"callback","title":"Sorcery.Schema.FieldType.get_sd_field/1","doc":"","ref":"Sorcery.Schema.FieldType.html#c:get_sd_field/1"},{"type":"callback","title":"Sorcery.Schema.FieldType.is_valid?/3","doc":"During validation of a field, we go through every attribute, comparing it to the given value to see if everything is valid.\n\nWe do it this way to try to keep everything related to attributes all in one place - the field struct.\nFor example, if you look in Sorcery.Schema.FieldType.Integer, you will find a defstruct that lists several attributes. \nBut there is also the is_valid? function, with a different head for each one.","ref":"Sorcery.Schema.FieldType.html#c:is_valid?/3"},{"type":"function","title":"Sorcery.Schema.FieldType.new/1","doc":"","ref":"Sorcery.Schema.FieldType.html#new/1"},{"type":"callback","title":"Sorcery.Schema.FieldType.new/1","doc":"","ref":"Sorcery.Schema.FieldType.html#c:new/1"},{"type":"function","title":"Sorcery.Schema.FieldType.new/2","doc":"","ref":"Sorcery.Schema.FieldType.html#new/2"},{"type":"behaviour","title":"Sorcery.StoreAdapter","doc":"A store adapter is a module that allows a PortalServer to access a data store by taking SrcQL, and converting it into something compatible with the specific store.\n\nCurrently there are only two Adapters available: Ecto, and InMemory.\nInMemory is used behind the scenes by the LiveHelper, and is basically a noop, preferring to keep all data in the portal itself.\nWhile the `Sorcery.StoreAdapter.Ecto` adapter is used for dealing with serious backends like MySql and Postgres.","ref":"Sorcery.StoreAdapter.html"},{"type":"function","title":"Sorcery.StoreAdapter.mutation/3","doc":"","ref":"Sorcery.StoreAdapter.html#mutation/3"},{"type":"function","title":"Sorcery.StoreAdapter.query/4","doc":"","ref":"Sorcery.StoreAdapter.html#query/4"},{"type":"callback","title":"Sorcery.StoreAdapter.run_mutation/2","doc":"","ref":"Sorcery.StoreAdapter.html#c:run_mutation/2"},{"type":"callback","title":"Sorcery.StoreAdapter.run_query/3","doc":"","ref":"Sorcery.StoreAdapter.html#c:run_query/3"},{"type":"module","title":"Sorcery.StoreAdapter.Ecto","doc":"This adapter requires the repo module to be passed in as args. \n\nHere is an example PortalServer\n\n```elixir\ndefmodule Src.PortalServers.Postgres do\n  use GenServer\n\n  def init(_) do\n    state = %{} # You can still add whatever you want here\n\n    state = Sorcery.PortalServer.add_portal_server_state(state, %{\n      config_module: Src,  # See below\n      store_adapter: Sorcery.StoreAdapter.Ecto,\n\n      args: %{\n        repo_module: MyApp.Repo\n      }\n    })\n    {:ok, state}\n  end\n\n\n  def handle_info({:sorcery, msg}, %{sorcery: inner_state} = outer_state) do\n    inner_state = Sorcery.PortalServer.handle_info(msg, inner_state)\n    {:noreply, Map.put(outer_state, :sorcery, inner_state)}\n  end\n\n\nend\n```","ref":"Sorcery.StoreAdapter.Ecto.html"},{"type":"module","title":"Sorcery.StoreAdapter.Ecto.Thread","doc":"Since we are translating SrcQL into SQL, there are some oddities.\nOne is how we handle completely unrelated lvars that are not joined.\nSo in this adapter we cluster a group of where clauses and joins together, and call it a thread.\nWe must call Repo.all(thread) seperately with each thread.","ref":"Sorcery.StoreAdapter.Ecto.Thread.html"},{"type":"function","title":"Sorcery.StoreAdapter.Ecto.Thread.all/1","doc":"","ref":"Sorcery.StoreAdapter.Ecto.Thread.html#all/1"},{"type":"function","title":"Sorcery.StoreAdapter.Ecto.Thread.all/3","doc":"","ref":"Sorcery.StoreAdapter.Ecto.Thread.html#all/3"},{"type":"function","title":"Sorcery.StoreAdapter.Ecto.Thread.new/2","doc":"","ref":"Sorcery.StoreAdapter.Ecto.Thread.html#new/2"},{"type":"function","title":"Sorcery.StoreAdapter.Ecto.Thread.put/3","doc":"","ref":"Sorcery.StoreAdapter.Ecto.Thread.html#put/3"},{"type":"exception","title":"Sorcery.NoPortalError","doc":"","ref":"Sorcery.NoPortalError.html"},{"type":"exception","title":"Sorcery.NoPortalInStateError","doc":"","ref":"Sorcery.NoPortalInStateError.html"},{"type":"extras","title":"README","doc":"","ref":"readme.html"},{"type":"extras","title":"This is not production ready. I am fixing bugs nearly daily right now. - README","doc":"# Sorcery\n\nA framework which rethinks how data flows, and how we build apps.\n\nPlays nicely with Phoenix LiveViews, or can be used alone.\n\nThe philosophy of Sorcery:\n\n> Suppose a database has a table called People.  \nThere are 3 different web pages that all display *some* data about the row where People.id == 1  \nNo matter where in the app this person appears, and no matter which columns are known about them, they are conceptually *the same entity*.  \nIf they change their name in one place, it should automatically update their name in EVERY place that it appears.  \nThis should be the DEFAULT way of building an app.  \nI don't want to setup a Phoenix.Channel, nor a PubSub.  \nI don't have time to wire this up in every page where I want it to happen.  \nIt should be automatic without thinking about it.  \nAll data should stay up to date, at all times, forever.  \nBut when I apply a mutation, it should happen in a functional, immutable way, with all the benefits one would expect.","ref":"readme.html#this-is-not-production-ready-i-am-fixing-bugs-nearly-daily-right-now"},{"type":"extras","title":"See the demo - README","doc":"These docs are horribly incomplete. But you can clone and run the demo app [https://github.com/greetingsfellowhumans/sorcery_demo.git](https://github.com/greetingsfellowhumans/sorcery_demo.git)\nIt is extremely minimalistic. No css whatsoever... but it shows how everything works together.","ref":"readme.html#see-the-demo"},{"type":"extras","title":"But how *does* it work?! - README","doc":"Sorcery comes with it's own headless, reversible, query language (SrcQL). By default it uses Ecto and whatever backend you want (Postgres, etc.). \nYou run the query and get a 'Portal' which is like a little wormhole showing whatever database entities match the Query.\n\nHere we see it in LiveViews, but any GenServer can do this.\nThe query lists several 'Logic Variables' or 'Lvars', and we are looking at one called \"?all_players\"\n\n```elixir\n~H\"\"\"\n  <%= for %{id: id, health: health, name: name} = _player <- portal_view(@sorcery, :my_portal, \"?all_players\") do %>\n     <%= id %> | <%= name %>'s health: <%= health %> \n  <% end %>\n\"\"\"\n```\n\nIf someone new enters the room, you'll see the change instantly. Ditto if someone leaves, or changes their name, or anything like that.\n\n\nSo a Portal is almost like a PubSub, if a PubSub had access to a full featured language, and the ability to recursively watch OTHER PubSubs.","ref":"readme.html#but-how-does-it-work"},{"type":"extras","title":"Performance - README","doc":"Memory hog with almost no latency.\n\nAll entities being watched are cached in mnesia tables. So this often allows us to skip calling the database.","ref":"readme.html#performance"},{"type":"extras","title":"Installation - README","doc":"```elixir\ndef deps do\n  [\n    {:sorcery, \"~> 0.3.8\"},\n  ]\nend\n```\n\nThen run the following to get bootstrapped\n```bash\n$ mix deps.get\n$ mix sorcery.init\n```\n\nSeveral files were just created in your app under /lib/src\n\nMake sure to also open your `application.ex` and add\n```elixir\nchildren = [\n  # Just an example, but you need to add each of your PortalServers here.\n  {Src.PortalServers.Postgres, name: Src.PortalServers.Postgres},\n\n  # This is mandatory.\n  {Src, []},\n]\n```\n\nThe next thing you should do is create some schemas, and queries.","ref":"readme.html#installation"},{"type":"extras","title":"Local development - README","doc":"If you want to work on Sorcery itself, use the bash environment variable SORCERY_DEVELOPMENT=true to get around some nasty conflicts.\n\n```bash\ngit clone https://github.com/greetingsfellowhumans/sorcery\ncd sorcery\nmix deps.get\nSORCERY_DEVELOPMENT=true mix test.watch\n```","ref":"readme.html#local-development"},{"type":"extras","title":"Changelog","doc":"# Changelog\nDates are in the format of yyyy-mm-dd\n\n## [Unreleased]\n- Better documentation.\n- Clean up the demo app\n\n## [0.4.11] - 2024-11-03","ref":"changelog.html"},{"type":"extras","title":"Fixed - Changelog","doc":"- When using LiveHelpers portal_view or portal_one, raises if searching for an lvar that doesn't exist in that portal.\n\n## [0.4.10] - 2024-11-01\nThis is a complete rewrite of the query engine for the Ecto adapter.\nInstead of manually building an entire query, we now leverage Ecto.Multi.\nIn SrcQL, every clause per Lvar will be put into its own Multi.all(...), and is given access to the previous results.\nJust as before, it is important that SrcQL clauses come in an order that makes sense. Do not reference an Lvar that appears later.\nThis change fixes several bugs with queries.","ref":"changelog.html#fixed"},{"type":"extras","title":"Fixed - Changelog","doc":"- Several SrcQL query bugs \n\n## [0.4.9] - 2024-09-22","ref":"changelog.html#fixed"},{"type":"extras","title":"Added - Changelog","doc":"- New field type :decimal, for whenever you are working with :numeric fields in ecto migrations.\n\n## [0.4.8] - 2024-09-15","ref":"changelog.html#added"},{"type":"extras","title":"Added - Changelog","doc":"- You can now put a handle_success/2 callback into spawn_portal\n\n## [0.4.7] - 2024-08-18","ref":"changelog.html#added"},{"type":"extras","title":"Fixed - Changelog","doc":"- Optimistic updates were not respecting lvar names\n\n## [0.4.6] - 2024-08-05","ref":"changelog.html#fixed"},{"type":"extras","title":"Fixed - Changelog","doc":"- table not found error, usually when app is starting up.\n- Fixed a bug with the reverse query when not all dependencies are met.\n- Fixed a bug where reverse query broke while deleting entities\n\n## [0.4.5] - 2024-08-01","ref":"changelog.html#fixed"},{"type":"extras","title":"Fixed - Changelog","doc":"Broken handle_success and handle_fail options for mutations","ref":"changelog.html#fixed"},{"type":"extras","title":"Changed - Changelog","doc":"handle_success and handle_fail options with LiveHelpers now update the socket instead of sorcery state.\n\n\n## [0.4.4] - 2024-07-29","ref":"changelog.html#changed"},{"type":"extras","title":"Fixed - Changelog","doc":"Broken call to put_flash\n\n## [0.4.3] - 2024-07-21","ref":"changelog.html#fixed"},{"type":"extras","title":"Added - Changelog","doc":"Mutation.send_mutation now takes an optional 3rd argument.\nThis must be a keyword list of options. \nThere are three options\n  :optimistic     | true | Whether to try to do an optimistic update. Does not work well when creating new entities.\n  :handle_fail    | nil | a function that takes the error, and the state. Returns state.\n  :handle_success | nil | a function that takes the error, and the state. Returns state.","ref":"changelog.html#added"},{"type":"extras","title":"Fixed - Changelog","doc":"Ecto Adapter error handling when a transaction fails.\n\n## [0.4.2] - 2024-07-06","ref":"changelog.html#fixed"},{"type":"extras","title":"Fixed - Changelog","doc":"- A bug that would crash mutations whenever the portal had an empty value for an lvar.\n\n## [0.4.1] - 2024-06-22","ref":"changelog.html#fixed"},{"type":"extras","title":"Fixed - Changelog","doc":"- Several bugs with the SrcQL Ecto Adapter\n- Add a :pending_portals list to inner_state, and a has_loaded? function to LiveHelpers.\n- A bug when you insert multiple entities and one depends on the other.\nNote there might still be an ecto foreign key constraint error. Dropping the constraint solved it for me, but I wish there were a cleaner solution.\n\n- Allow spawn_portal to work even in an environment without the html helpers\n\n- Fix a bug in which only one portal per pid could ever be found by the reverse query","ref":"changelog.html#fixed"},{"type":"extras","title":"Breaking Changes - Changelog","doc":"When creating new entities, it will no longer autofill missing data with randomly generated content.","ref":"changelog.html#breaking-changes"},{"type":"extras","title":"Added - Changelog","doc":"- SrcQL Ecto adapter now handles nil comparison safely.\n- Mutation.skip/3\n- Mutation.validate/3\n- LiveHelpers add an optimistic_mutations function\n\n## [0.4.0] - 2024-06-22","ref":"changelog.html#added"},{"type":"extras","title":"Fixed - Changelog","doc":"A lot of the automatic data generation was broken. Rebuilt much of it, added some pretty thorough prop tests.","ref":"changelog.html#fixed"},{"type":"extras","title":"Breaking Changes - Changelog","doc":"The Schema defaults are somewhat different, so you should be aware that:\n- List and map fields now default to optional?: false because it makes more sense for an empty list to be `[]` than `nil`\n\n\n## [0.3.7] - 2024-06-19","ref":"changelog.html#breaking-changes"},{"type":"extras","title":"Fixed - Changelog","doc":"- Typespec warnings\n- Issues when Phoenix mounts twice, particularly the first time when the expected data is not present\n\n## [0.3.6] - 2024-06-19","ref":"changelog.html#fixed"},{"type":"extras","title":"Fixed - Changelog","doc":"- Conflict when starting tests on an app that depends on sorcery.\nworkaround for now is setting an environment varialbe SORCERY_DEVELOPMENT=true when working on sorcery itself\n- Cleaned up several warnings\n\n## [0.3.5] - 2024-06-18","ref":"changelog.html#fixed"},{"type":"extras","title":"Fixed - Changelog","doc":"Mutation.create_entity was broken. I fixed it.\n\n## [0.3.4] - 2024-06-17","ref":"changelog.html#fixed"},{"type":"extras","title":"Added - Changelog","doc":"- New field types for schemas: float, boolean, list, map\nNote that list fields must contain a :coll_of type. for example:\n```elixir\nmy_items: %{t: :list, coll_of: :integer}\n```\n\nWhile using ecto/postgres, lists and maps are stored as json. I have not tested this with anything else\n\n## [0.3.3] - 2024-06-17","ref":"changelog.html#added"},{"type":"extras","title":"Added - Changelog","doc":"- The Killbot module\n  This is basically a garbage collection system. Since SorceryDb stores a lot of entities in :mnesia, and some other portal data in :ets, Killbot is a process that periodically checks which entities are no longer being used anywhere, and safely removes them.","ref":"changelog.html#added"},{"type":"extras","title":"Changed - Changelog","doc":"New options for your Src module\n```elixir\nuse Sorcery,\n    killbot: %{\n      interval: int,\n      postmortem_delay: int\n    }\n```\nThese are optional. \nIn theory, Killbot should not require any setup, it is initialized by the `use Sorcery` under the hood.","ref":"changelog.html#changed"},{"type":"extras","title":"Breaking Changes - Changelog","doc":"None (in theory)\n\n\n## [0.3.2] - 2024-06-01","ref":"changelog.html#breaking-changes"},{"type":"extras","title":"Added - Changelog","doc":"- Optimistic Updates\n  Now when you send a mutation, it will return a :temp_data field, guessing at the changes. This allows us to see the updated data long before the changes reach the parent portal_server or database.\n  temp_data is reset after the real data comes in\n  Mutations cannot be run on a portal that already has temp_data. That would be too complex. Extra mutations are silently dropped.\n- More documentation, especially for Sorcery.Query","ref":"changelog.html#added"},{"type":"extras","title":"Changed - Changelog","doc":"- %Sorcery.PortalServer.InnerState{}\n  This has eased a lot of confusion and unveiled bugs waiting to happen.\n- Turned Sorcery.Query into a behaviour, and the __using__ macro into an implementation of it.\n    Now in exdoc, all the functions defined by Use are well documented with examples that pass tests.","ref":"changelog.html#changed"},{"type":"extras","title":"live_views","doc":"","ref":"live_views.html"},{"type":"extras","title":"LiveViews - live_views","doc":"Let's start with the minimalistic demo module for a LiveView, and explain it afterward\n\n```elixir\ndefmodule MyAppWeb.SandboxLive do\n  use Phoenix.LiveView\n  use Sorcery.LiveHelpers  # <- This adds functions like spawn_portal/2 etc.\n\n  def mount(_param, _sesh, socket) do\n    socket = \n      socket\n      |> initialize_sorcery(%{\n        sorcery_module: MyApp.Sorcery # We talk about this module in the Introduction guide\n      })\n\n      # You can call this multiple times with different config\n      # As long as the portal_name is unique on a per-liveview basis.\n      |> spawn_portal(%{\n        portal_name: :battle_portal, # whatever atom you want\n        portal_server: MyApp.Sorcery.PortalServers.Postgres, # See the PortalServer guide\n        query_module: MyApp.Sorcery.Queries.GetBattle, # See the Queries guide\n        query_args: %{player_id: 1} # this will be passed into the query eventually\n      })\n\n    {:ok, socket}\n  end\n\n\n  def handle_event(\"change_hp\", %{\"id\" => idstr, \"amount\" => amountstr}, socket) do\n    id = String.to_integer(idstr)\n    amount = String.to_integer(amountstr)\n\n    # Every mutation requires a portal\n    Sorcery.Mutation.init(socket.assigns.sorcery, :battle_portal)\n    |> Sorcery.Mutation.update([:player, id, :health], fn _old_health, health -> health + amount end)\n    |> Sorcery.Mutation.send_mutation()\n        \n    {:noreply, socket}\n  end\n\n\n  def render(assigns) do\n    ~H\"\"\"\n      <%= for %{id: id, health: health, name: name} = _player <- portal_view(@sorcery, :battle_portal, \"?all_players\") do %>\n         <%= id %> | <%= name %>'s health: <%= health %> \n         Heal  \n         Harm  \n\n      <% end %>\n    \"\"\"\n  end\n\n\nend\n```","ref":"live_views.html#liveviews"},{"type":"extras","title":"Configuration","doc":"# Configuration\n\nHerein lies the key to writing a schema. Here is how we would rewrite [the Example given by Ecto.Schema](https://hexdocs.pm/ecto/Ecto.Schema.html#module-example):\n\n```elixir\ndefmodule App.User do\n#######\n# We don't need this anymore\n#######\n#\n#use Ecto.Schema\n#\n#schema \"users\" do\n#  field :name, :string\n#  field :age, :integer, default: 0\n#  field :password, :string, redact: true\n#  has_many :posts, Post\n#end\n\n\n  use Sorcery.Schema, %{\n    meta: %{},\n    fields: %{\n        name:      %{t: :string},\n        age:       %{t: :integer, default: 0},\n        password:  %{t: :string,  redact: true},\n        # No need for a has_many. In Sorcery, it is always the child that references the parent.\n    }\n  }\n\n\nend\n```\n\nFor a more complete example, look at this:\n```elixir\ndefmodule Src.Schemas.Types do\n  use Sorcery.Schema, fields: %{\n    bool_mandatory: %{t: :boolean, optional?: false},\n    bool_optional: %{t: :boolean, optional?: true},\n    int_mandatory: %{t: :integer, optional?: false, min: 0, max: 10},\n    int_optional: %{t: :integer, optional?: true, min: -10, max: 10},\n    float_mandatory: %{t: :float, optional?: false, min: 0.0, max: 10.0},\n    float_optional: %{t: :float, optional?: true, min: -10.0, max: 10.0},\n    string_mandatory: %{t: :string, optional?: false, min: 10, max: 20},\n    string_optional: %{t: :string, optional?: true, min: 10, max: 20},\n    list_mandatory: %{t: :list, coll_of: :integer, min: 10, max: 20},\n    list_optional:  %{t: :list, coll_of: :string, optional?: true,  min: 10, max: 20},\n    list_inner: %{t: :list, coll_of: :string, inner: %{min: 15, max: 25}, min: 5, max: 10},\n    map_mandatory: %{t: :map, optional?: false},\n    map_optional: %{t: :map, optional?: true},\n  }\nend\n```","ref":"configuration-4.html"},{"type":"extras","title":"Optional? - Configuration","doc":"In the above example, notice how we explicitly declare the :optional? key in every field. If this is set to true, and if no value is provided while generating the struct, then the field will use the FieldType default. For most fields (:integers, :strings, etc) this is nil. For lists and maps, it's an empty list or map respectively.\nFor example an :integer field with optional?: true will only ever be an integer.\nBut an :integer field with optional?: false might be either an integer or nil.","ref":"configuration-4.html#optional"},{"type":"extras","title":"Meta Map - Configuration","doc":"Before getting into the meat of it, notice the empty map called :meta up there? We can use it to set some schema-wide configuration.\n\n| :meta         | types           | Default       | Description      |\n| ---           | ---             | ---           | ---              |\n| :optional?    | boolean         | true          | Whether every field allows nil. Individual fields can override. Maps and lists default to false. |","ref":"configuration-4.html#meta-map"},{"type":"extras","title":"Fields by type - Configuration","doc":"### :boolean\n|               | types           | Optional? | Default | Description      |\n| ---           | ---             | ---       | ---     | ---              |\n| :default      | boolean         | true      | nil     | If no value is given, this will be used.          |\n| :optional?    | boolean         | true      | nil     | Whether field is allowed to be nil      |\n\n\n### :string\n|               | types           | Optional? | Default | Description      |\n| ---           | ---             | ---       | ---     | ---              |\n| :min          | integer         | true      | 0       | field must have at least this many characters.     |\n| :max          | integer         | true      | 25      | field must have no more than this many characters. |\n| :default      | string          | true      | nil     | If no value is given, this will be used.          |\n| :optional?    | boolean         | true      | true    | Whether field is allowed to be nil      |\n\n\n### :integer\n|               | types           | Optional? | Default | Description                     |\n| ---           | ---             | ---       | ---     | ---                             |\n| :min          | integer         | true      | -10_000 | field be at least this big.     |\n| :max          | integer         | true      | 10_000  | field be no more than this big. |\n| :default      | integer         | true      | nil     | If no value is given, this will be used.          |\n| :optional?    | boolean         | true      | true    | Whether field is allowed to be nil      |\n\n### :float\n|               | types           | Optional? | Default   | Description                     |\n| ---           | ---             | ---       | ---       | ---                             |\n| :min          | float           | true      | -10_000.0 | field be at least this big.     |\n| :max          | float           | true      | 10_000.0  | field be no more than this big. |\n| :default      | integer         | true      | nil       | If no value is given, this will be used.          |\n| :optional?    | boolean         | true      | true      |  Whether field is allowed to be nil      |\n\n### :list\n|               | types           | Optional? | Default   | Description                     |\n| ---           | ---             | ---       | ---       | ---                             |\n| :min          | integer         | true      | 0         | list must have at least this many items     |\n| :max          | integer         | true      | 10        | list must have at most this many items     |\n| :default      | integer         | true      | nil       | If no value is given (and optional?: true), then nil.          |\n| :optional?    | boolean         | true      | false     | Whether field is allowed to be nil      |\n\n### :map\n|               | types           | Optional? | Default   | Description                     |\n| ---           | ---             | ---       | ---       | ---                             |\n| :default      | integer         | true      | nil       | If no value is given (and optional?: true), then nil.          |\n| :optional?    | boolean         | true      | false     | Whether field is allowed to be nil      |\n\nWe don't both generating content inside maps. Too much complexity, and too far from the core goals of Sorcery.\n\n### :fk\nThis is still under construction. For now I would suggest using :integer instead.","ref":"configuration-4.html#fields-by-type"},{"type":"extras","title":"Introduction","doc":"# Introduction\n\nSorcery Query Language is a reversible query language. Please don't call it SQL, that would be confusing for some reason. SrcQL will do fine.\n\nThe syntax is partially inspired by Datalog, which is used by the Datomic database. But there are many important differences, and it is not intended to be identical.","ref":"introduction-2.html"},{"type":"extras","title":"Basic Usage - Introduction","doc":"As a quirky example, we are making a multiplayer game in which there are players, organized into teams, and each player is a wizard who casts spells. They can all battle in an 'arena'\nSo the four tables are: team, player, spell, arena\n\nNow suppose we want a query that pulls up data for arena 1 \n\n```elixir\n# Don't try to fully understand this yet. We're going to break it down soon and explain it line by line.\n\nalias Sorcery.Query, as: SrcQL\nSrcQL.new(%{\n  where: [\n    [\"?arena\",   :arena,   :id, :args_arena_id],\n    [\"?teams\",   :team,    :current_arena_id, \"?arena.id\"],\n    [\"?players\", :player, [\n      {:team_id, \"?teams.id\"},\n      {:health, {:>, 0}}\n    ]],\n    [\"?spells\",  :spell,   :player_id, \"?players.id\"]\n  ],\n  args: %{\n    arena_id: :integer\n  },\n  find: %{\n    \"?players\" => :*,\n    \"?spells\" =>  [:name, :energy, :type, :player_id]\n  }\n})\n\nargs = %{arena_id: 1}\n\n\n# Now when you assign this Query to a Portal called :portal_name, and you pass in the args above, you can forevermore view the \"?players\" and \"?spells\" lvars\niex> portal_view(@sorcery, :portal_name, \"?players\")\n[\n  %{id: 2, name: \"Jose\", health: 100, team_id: 1},\n  %{id: 85, name: \"Aaron\", health: 10, team_id: 2},\n]\n```\n\n\nSo you might have guessed the list of :where tuples is similar to something you might find in SQL. Just a list of conditions for filtering data. Except the tuples look weeeeiiird.\n\n```elixir\n  # each tuple has 4 elements in it.\n  # {lvar,     tk,       attr,  value}\n  {\"?arena\",   :arena,   :id,   1}, \n```\nThe string `\"?arena\"` is a logic variable, or Lvar for short.\nit represents a lazy Set of entities of a specific type. in this case they are all :arena entities.\nThe attr/value pair helps us limit the set of matches. Now every entity in the ?arena set must have an :id of 1. \nSo how big is the set going to be? Either 0 or 1.\n\nLet's continue.\n\n```elixir\n  # Now the final element (the value) is referencing a previous lvar... and appending \".id\" to the end of it.\n  # If you imagine gettings all the possible ids from ?arena, i.e. ids = Enum.map(?arena, &(&1.id))\n  # And then we are basically running Enum.filter(?teams, &(&1.current_arena_id in ids))\n  {\"?teams\",   :team,    :current_arena_id, \"?arena.id\"}\n```\nAh yes, our first inferred join. Instead of a JOIN...ON syntax, we just do this and it works like ~~magic~~ sorcery.\n\n\n```elixir\n  # Boring, you know this one already.\n  {\"?players\", :player,  :team_id, \"?teams.id\"}\n\n  # Ahh now we bring operators into the mix!\n  # So we only get players who are members of at least one of the ?teams...\n  # But they ALSO must not be dead. Makes it hard to play. So it filters player.health > 0\n  {\"?players\", :player,  :health, {:>, 0}}\n```\n\nBut what about :find? Well it's kind of like if a SQL select statement made a baby with Map.take/2.\nNote that it automatically adds an :id field to every take... and the format returned is a map in a pretty specific shape.\n```elixir\nfind: %{\"?foo\" => [:name] }\n\n# lvar => %{id => entity}\n\nselection = %{\n    \"?foo\" => %{\n        1 => %{id: 1, name: \"some name\"}\n    }\n}\n```\n\n@TODO have the option to exclude :id fields.\n\n@TODO have the option to use a different primary key than :id.","ref":"introduction-2.html#basic-usage"},{"type":"extras","title":"Function syntax - Introduction","doc":"What if I told you the :where tuples were actually just a shorthand for a more detailed syntax?\n\nLet's rewrite the query\n\n\n```elixir\n\nalias Sorcery.Query, as: SrcQL\nSrcQL.new(%{\n  where: [\n    SrcQL.Clause.new(%{lvar: \"?arena\", tk: :arena, filters: [%{attr: :id, val: 1, op: :==}]}),\n    SrcQL.Clause.new(%{lvar: \"?teams\", tk: :team,  filters: [%{attr: :team_id, lvar: \"?arena\", lvar_attr: \"id\", op: :in}] }),\n    SrcQL.Clause.new(%{lvar: \"?players\", tk: :player, filters: [\n      %{attr: :team_id, lvar: \"?arena\", lvar_attr: \"id\", op: :in},\n      %{attr: :health, val: 0, op: :>}\n    ]}),\n    SrcQL.Clause.new(%{lvar: \"?spells\", tk: :spell, filters: [%{attr: :player_id, lvar: \"?players\", lvar_attr: \"id\", op: :in}]}),\n  ],\n  find: %{\n    \"?players\" => [:name, :health, :team_id]\n    \"?spells\" =>  [:name, :energy, :type, :player_id]\n  }\n})\n```\nThese functions return structs %SrcQL.Clause{...}\nAs do the tuples. It's all the same.\n\nNow why on earth would you want this? Well you probably don't want to do it manually. But it's useful to know, because there are also other structs available, for different types of clauses!\n\n```elixir\n# This will get all foo entities with :magic greater than 100, OR with an id of 42.\nSrcQL.OrClause.new(%{lvar: \"?foo\", tk: :foo, filters: [\n  %{attr: :id, val: 42},\n  %{attr: :magic, val: 100, op: :>},\n]})\n```\nThere is no need for an AndClause, because the default Clause does that.\n\nYou can also do certain aggregation operations.\n\n```elixir\n# This will get the average health of all entities matching \"?players\"\nSrcQL.Aggregate.Mean.new(%{lvar: \"?avg_health\", val: \"?players.health\"})\n```","ref":"introduction-2.html#function-syntax"},{"type":"extras","title":"Introduction","doc":"# Introduction\n\nPortals are what tie it all together. The connective tissue of Sorcery. But first we have one other concept to talk about.","ref":"introduction-1.html"},{"type":"extras","title":"PortalServers - Introduction","doc":"A PortalServer is an elixir process which has access to a data store. When you send a SorceryQuery to a PortalServer, you don't *just* get some data back... you create a Portal, with a continuous feed of data matching that query. Kind of like a PubSub on steroids.\n\nThe PortalServer holds all the configuration, describing in detail HOW the data is accessed. \n\nHere you can throw in your own custom adapters for different databases. \n\nYou can tell it to pass messages via REST and GraphQL or elixir's Process.send\n\nAs long as you find/build the right adapters, the options are limitless.","ref":"introduction-1.html#portalservers"},{"type":"extras","title":"Fun examples of PortalServers to get your brain juices flowing: - Introduction","doc":"| PortalServer      | Data Store                  |\n| ---               |  ---                        |\n| Phoenix.LiveView  |  assigns.sorcery            |\n| GenServer         |  a map in the state         |\n| GenServer         |  Repo connected to Postgres |\n| GenServer         |  ANOTHER PortalServer       |\n| A plain Process   |  An external weather API    |\n\nDo you see how flexible it is? Very. I believe one could technically have a PortalServer fire off an email, which someone manually responds to with an attachment... as long as you write an adapter for it, and don't care about latency, it would work.","ref":"introduction-1.html#fun-examples-of-portalservers-to-get-your-brain-juices-flowing"},{"type":"extras","title":"LiveView - Introduction","doc":"This gets special mention because it was the original use case, and reason for creating Sorcery.\n\n```elixir\n  def mount(_, _, socket) do\n    socket = socket\n             |> Sorcery.PortalServer.create()\n             |> Sorcery.Portal.open(to, sorcery_query)\n    {:ok, socket}\n  end\n\n  def handle_info({:sorcery, msg}, socket), do: Sorcery.PortalServer.live_view_handler(msg, socket)\n```\n\nThat is all we need to do to turn a LiveView into a PortalServer. And just for fun, we also open up a Portal to some other PortalServer. See the guides on Queries and LiveViews for more detail.\n\nNote that there will be times when this LiveView receives all sorts of whacky messages from whatever other PortalServers it connects to. You don't need to care about those whatsoever, they will all be tagged as {:sorcery, _}\n\nNow you can obviously intercept those and look at them with IO.inspect, or dbg, just for fun. But you do not need to know what's going on behind the curtain. I am a powerful wizard, you can trust me.","ref":"introduction-1.html#liveview"},{"type":"extras","title":"Configuration","doc":"# Configuration\n\nThe cheatsheet for building a PortalServer\n\n```elixir\ndefmodule Src.PortalServers.Demo do\n  use GenServer\n\n  @config %{\n    store_args: %{repo: MyApp.Repo},\n    store_plugin: Sorcery.Plugins.Ecto\n  }\n\n  def init() do\n  end\n\n\nend\n```\n\n\n| t: :string    | types           | Optional? | Default                    | Description      |\n| ---           | ---             | ---       | ---                        | ---              |\n| :prune_dead   | integer         | false     | 60_000                     | How often (ms) do we remove queries from dead pids. |\n| :store_plugin | module          | false     | Sorcery.Plugins.GenServer  | Module for reading and writing to storage           |\n| :store_args   | map             | false     | %{}                        | Any extra data needed by the store_plugin goes here |","ref":"configuration-3.html"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,51,46,48]}}